var Readable = require('stream').Readable;
var fs = require('fs');
var AWS = require('hws-sdk');
var util = require('util');
var pool;
var kMinPoolSpace = 128;

var CLRF = "\r\n";
var  CHUNK_STRING_TO_SIGN_PREFIX = "AWS4-HMAC-SHA256-PAYLOAD";
var  CHUNK_SIGNATURE_HEADER = ";chunk-signature=";
var SIGNATURE_LENGTH = 64;
var DEFAULT_CHUNK_SIZE=64*1024;


var FINAL_CHUNK = new Buffer(0);



function allocNewPool(poolSize) {
    pool = new Buffer(poolSize);
    pool.used = 0;
}

function calculateSignedChunkLength(chunkDataSize) {

    var len = chunkDataSize.toString('16').length
        + CHUNK_SIGNATURE_HEADER.length
        + SIGNATURE_LENGTH
        + CLRF.length
        + chunkDataSize
        + CLRF.length;
    return len;
}
util.inherits(ChunkedReadStream, Readable);
fs.createChunkedReadStream = function(path, options) {
    return new ChunkedReadStream(path, options);
};
function ChunkedReadStream(path,options){
    options = util._extend({
        highWaterMark: 65626
    }, options || {});

    Readable.call(this, options);

    this.dateTime=options.dateTime;
    this.keyPath=options.keyPath;
    this.headerSignature='';
    this.signingKey='';
    this.priorChunkSignature='';
    this.path = path;

    this.fd = options.hasOwnProperty('fd') ? options.fd : null;
    this.flags = options.hasOwnProperty('flags') ? options.flags : 'r';
    this.mode = options.hasOwnProperty('mode') ? options.mode : 438; /*=0666*/

    this.start = options.hasOwnProperty('start') ? options.start : undefined;
    this.end = options.hasOwnProperty('end') ? options.end : undefined;
    this.autoClose = options.hasOwnProperty('autoClose') ?
        options.autoClose : true;
    this.pos = undefined;
    this.ended=false;
    if (!util.isUndefined(this.start)) {
        if (!util.isNumber(this.start)) {
            throw TypeError('start must be a Number');
        }
        if (util.isUndefined(this.end)) {
            this.end = Infinity;
        } else if (!util.isNumber(this.end)) {
            throw TypeError('end must be a Number');
        }

        if (this.start > this.end) {
            throw new Error('start must be <= end');
        }

        this.pos = this.start;
    }

    if (!util.isNumber(this.fd))
        this.open();
    var self=this;

    this.on('end', function() {
        this.priorChunkSignature='';
        if (this.autoClose) {
            this.destroy();
        }
    });
};

fs.ChunkedReadStream = ChunkedReadStream;
ChunkedReadStream.prototype.open = function() {
    var self = this;
    fs.open(this.path, this.flags, this.mode, function(er, fd) {
        if (er) {
            if (self.autoClose) {
                self.destroy();
            }
            self.emit('error', er);
            return;
        }

        self.fd = fd;
        self.emit('open', fd);
        // start the flow of data.
        self.read();
    });
};
ChunkedReadStream.prototype.stringToSign= function(chunk) {
    var parts = [];
    parts.push(CHUNK_STRING_TO_SIGN_PREFIX);

    parts.push(this.dateTime);
    parts.push(this.keyPath);
    parts.push(this.priorChunkSignature);

    parts.push(AWS.util.crypto.sha256('', 'hex'));
    parts.push(AWS.util.crypto.sha256(chunk, 'hex'));
    return parts.join('\n');
};

ChunkedReadStream.prototype._read = function(n) {
    if (!util.isNumber(this.fd)){
        return this.once('open', function() {
            this._read(n);
        });
    }
    if(this.ended){
        this.push(null);
        return;
    }
    if (this.destroyed){
        return;
    }


    if (!pool || pool.length - pool.used < kMinPoolSpace) {
        // discard the old pool.
        pool = null;
        allocNewPool(DEFAULT_CHUNK_SIZE);
    }

    // Grab another reference to the pool in the case that while we're
    // in the thread pool another read() finishes up the pool, and
    // allocates a new one.
    var thisPool = pool;
    var toRead = Math.min(pool.length - pool.used, n);
    var start = pool.used;

    if (!util.isUndefined(this.pos))
        toRead = Math.min(this.end - this.pos + 1, toRead);

    // already read everything we were supposed to read!
    // treat as EOF.
    if (toRead <= 0){
        console.warn('EOF--->');
        return this.push(null);
    }
    // the actual read.
    var self = this;
    fs.read(this.fd, pool, pool.used, toRead, this.pos, onread);

    // move the pool positions, and internal position for reading.
    if (!util.isUndefined(this.pos)){
        this.pos += toRead;
    }
    pool.used += toRead;

    function onread(er, bytesRead) {
        if (er) {
            if (self.autoClose) {
                self.destroy();
            }
            console.error('fs read---->',er);
            self.emit('error', er);
        } else {
            var b = null;
            if (bytesRead > 0){
                b = thisPool.slice(start, start + bytesRead);
               var signature =AWS.util.crypto.hmac(self.signingKey, self.stringToSign(b), 'hex');

                var header=new Buffer(bytesRead.toString(16)+CHUNK_SIGNATURE_HEADER+signature+CLRF);
                var chunk=Buffer.concat([header,b,new Buffer(CLRF)]);
                self.push(chunk);
                self.priorChunkSignature=signature;

            }
            else{

                bytesRead=0;
                var signature =AWS.util.crypto.hmac(self.signingKey, self.stringToSign(FINAL_CHUNK), 'hex');
                var header=new Buffer(bytesRead.toString(16)+CHUNK_SIGNATURE_HEADER+signature+CLRF);
                var chunk=Buffer.concat([header,FINAL_CHUNK,new Buffer(CLRF)]);
                try{
                    self.push(chunk);
                    self.ended=true;
                }
                catch(e){
                    console.error(e);
                }
            }

        }
    }
};


ChunkedReadStream.prototype.destroy = function() {
    if (this.destroyed)
        return;
    this.destroyed = true;
    if (util.isNumber(this.fd))
        this.close();
};


ChunkedReadStream.prototype.close = function(cb) {
    var self = this;
    if (cb)
        this.once('close', cb);
    if (this.closed || !util.isNumber(this.fd)) {
        if (!util.isNumber(this.fd)) {
            this.once('open', close);
            return;
        }
        return process.nextTick(this.emit.bind(this, 'close'));
    }
    this.closed = true;
    close();

    function close(fd) {
        fs.close(fd || self.fd, function(er) {
            if (er)
                self.emit('error', er);
            else
                self.emit('close');
        });
        self.fd = null;
    }
};


ChunkedReadStream.prototype.calculateStreamContentLength=function(originalLength) {
    var maxSizeChunks = Math.round(originalLength / DEFAULT_CHUNK_SIZE);
    var remainingBytes =  originalLength % DEFAULT_CHUNK_SIZE;
    return maxSizeChunks * calculateSignedChunkLength(DEFAULT_CHUNK_SIZE)
        + (remainingBytes > 0? calculateSignedChunkLength(remainingBytes) : 0)
        + calculateSignedChunkLength(0);


}
