var AWS = require('hws-sdk/lib/core');
var fs = require('fs');
var path = require('path');
var db = require('./mongodb');
var DownloadPool = require("./DownloadPool");
var UploadBucketPool = require("./UploadBucketPool");
var obsConfig = require("./obsConfig.js");
var Task = require("./TaskModel.js").Task;

function TaskManager() {

}

constructor: function TaskManager(userId, s3, taskNum, uploadConfig, downloadConfig) {
    var self = this;

    AWS.SequentialExecutor.call(self);
    this.userId = userId;
    this.s3 = s3;

    if (!uploadConfig) {
        uploadConfig = new obsConfig.UploadConfig();
    }
    self.uploadConfig = uploadConfig;

    if (!downloadConfig) {
        downloadConfig = new obsConfig.DownloadConfig();
    }
    self.downloadConfig = downloadConfig;
    self.maxTaskNum = uploadConfig.maxUploadTasks;
    logger.info("max task num:", self.maxTaskNum);
    this.pauseAll = false;
    this.abortAll = false;

    this.runningMap = {};//所有任务运行集合 map{id:task}

    this.downloadPool = null; //下载池
    this.uploadBucketPool = null; //上传池

    this.enableRunningTask = true;  //是否允许执行任务

    db.getCollection(function (err, data) {
        if (data) {
            var queryString = {userId: self.userId, "stat": "RUNNING"};
            var param = {$set: {stat: "PAUSE"}};
            //如果上次退出异常关闭，修正数据库中RUNNING==>>PAUSE
            db.updateProperties(queryString, param, function (err, data) {
                self.createDownloadPool();
                self.createUploadPool();

                self.emit("dbconn");//数据连接成功
            });
        }
        if (err) {
            global.logger.error('Collection Mongodb Err:', err);
        }
    });
}

TaskManager.prototype.disableRunTask = function () {
    this.enableRunningTask = false;
}

TaskManager.prototype.getRunningMapSize = function () {
    return Object.getOwnPropertyNames(this.runningMap).length;
}

TaskManager.prototype.runTask = function () {
    var self = this;
    if (!self.enableRunningTask) {
        logger.info("[taskmanager] the taskmanage has been disabled.");
        return;
    }
    if (self.pauseAll) {
        logger.info("[taskmanager] the pauseall has been enable.");
        return;
    }
    if (self.getRunningMapSize() >= self.maxTaskNum) {
        return;
    }
    var queryString = {userId: self.userId, stat: "WAIT"};
    db.getAndModify(queryString, function (err, taskRecord) {  //如果本操作耗时太长，有可能导致多余maxTaskNum的task数量，但是是暂时的。
        if (!err && taskRecord && taskRecord.value) {
            var task = new Task(taskRecord.value);
            global.logger.info("execute task:", task.bucket, " ", task.key, task.stat, task.type);
            task.stat = "RUNNING";
            self.runningMap[task.id] = task;
            self.emit("onStart", [task]);
            if (task.type === Task.TaskTypeEnum.GET || task.type === "recoverAndDownload") {
                self.downloadPool.run(task);
            }
            else {
                self.uploadBucketPool.run(task);
            }
        } else {
            logger.info("[taskmanager] nothing get and modify from db.", err);
        }
    })

}

//插入任务到数据库
TaskManager.prototype.push = function push(task) {
    var self = this;
    if (!task) {
        return;
    }
    task.userId = self.userId;
    if (self.abortAll) {
        return;
    }
    if (self.pauseAll && (task.type !== 'recover' || task.type !== 'recoverAndDownload')) {
        task.stat = "PAUSE";
    }
    if (task.type === Task.TaskTypeEnum.GET || task.type === 'recoverAndDownload') {
        task.partSize = self.downloadConfig.multiDownloadNum;
    } else {
        task.partSize = self.uploadConfig.multiUploadNum;
    }
    //写入数据库
    db.add(task, function (err, data) {
        if (err) {
            global.logger.error("add task failed:", task, err);
            return;
        }
        self.runTask();
    });
};

//取消
TaskManager.prototype.abort = function abort(id) {
    var self = this;
    var task = self.runningMap[id];
    if (task && task.stat === "RUNNING") {
        global.logger.info("abort task:", task.bucket, " ", task.key, " ", task.type);
        //上传
        if (isInUploadPool(task)) {
            self.uploadBucketPool.abort(task);
        }
        //下载
        else if (task.type === "getObj" || (task.type === 'recoverAndDownload' && task.stat != "RESTORING")) {
            //取消正在执行的req
            self.downloadPool.abort(task);
        }
    }
    else {
        db.getAndRemove({id: id}, function (err, data) {
            if (!err) {
                task = data;
                if (task && task.stat === Task.TaskStatEnum.SUCCESS) {
                    return;
                }
                self.emit("onAbort", [task]);
            }
        });
    }
};

//暂停
TaskManager.prototype.pause = function pause(id, callback) {
    var self = this;
    var task = self.runningMap[id];
    if (task) {
        global.logger.info("pause task:", task.bucket, " ", task.key, " ", task.type, task.stat);
        if (task.stat === "RUNNING") {
            //上传
            if (isInUploadPool(task)) {
                self.uploadBucketPool.pause(task);
            }
            //下载
            else if (task.type === "getObj" || task.type === 'recoverAndDownload') {
                self.downloadPool.pause(task);
            }
        }
    }
    else {
        if(task.stat === "RESTORING"){
            return false;
        }
        db.get({id: id}, function (err, data) {
            if (!err) {
                task = data;
                logger.info("[PAUSE] TASK(not running):", task.key, task.stat);
                if (task.stat == "SUCCESS") {
                    return;
                }
                task.stat = "PAUSE";
                delete task['active'];
                db.updateStat(task, function (err, data) {
                    if (callback) {
                        callback.call();
                    }
                });
            }
        });
    }
};

//开始
TaskManager.prototype.start = function start(id) {
    var self = this;
    self.setPauseAll(false);
    db.get({id: id}, function (err, data) {
        if (!err)
            var task = data;
        if (task) {
            if (task.stat === "PAUSE" || task.stat === "FAILED" || (task.stat === "WAIT" && task.type === "recoverAndDownload")) {
                task.stat = "WAIT";
                db.updateStat(task, function (err, data) {
                    if (!err) {
                        self.emit("onWait", [task]);
                        self.runTask();
                    }
                })
            }
        }
    });

};

//开始所有
TaskManager.prototype.startAll = function (callback) {
    var self = this;
    self.setPauseAll(false);
    self.setAbortAll(false);
    var queryString = {userId: self.userId, type: {"$ne": "recover"}, "$or": [{"stat": "PAUSE"}, {"stat": "FAILED"}]};
    var param = {$set: {stat: "WAIT"}};
    db.updateProperties(queryString, param, function (err, data) {
        if (callback)
            callback(err, data);
    });
    setTimeout(() => {
        for (let index = 0; index < self.maxTaskNum; index++) {
            self.runTask();
        }
    }, 100);
};
//暂停所有
TaskManager.prototype.pauseAllTask = function (callback) {
    var self = this;
    self.setPauseAll(true);
    global.logger.info("pause all tasks!");
    var queryString = {userId: self.userId, "stat": "WAIT"};
    var param = {$set: {stat: "PAUSE"}};
    db.updateProperties(queryString, param, function (err, data) {
        if (callback)
            callback(err, data);
    });
};

//取消所有
TaskManager.prototype.abortAllTasks = function (callback) {
    var self = this;
    var query = {userId: self.userId, type: "getObj", stat: {$nin: ['SUCCESS', 'RESTORING', "MD5SUCCESS"]}};
    global.logger.info("Abort all tasks!");
    db.lists(query, function (err, data) {
        if (!err) {
            for (var i in data) {
                var filePath = path.normalize(data[i].path);
                fs.exists(filePath, function (isExists) {
                    if (isExists) {
                        fs.unlink(filePath, (err) => {
                        });
                    }
                });
            }
        } else {
            global.logger.error('lists download tasks error', err);
        }
        var queryString = {userId: self.userId, stat: {$nin: ['SUCCESS', 'RESTORING', "MD5SUCCESS"]}};
        db.delByQueryString(queryString, function (err, data) {
            if (callback)
                callback(err, data);
        });
    });
};

TaskManager.prototype.setPauseAll = function (flag) {
    this.pauseAll = flag;
};

TaskManager.prototype.setAbortAll = function (flag) {
    this.abortAll = flag;
};

//创建或回收buckepool
TaskManager.prototype.createUploadPool = function createUploadPool() {
    var self = this;
    self.uploadBucketPool = new UploadBucketPool(self.s3, db)
        .on("progress", function (task, info) {
            self.emit("progress", [task, info]);
        }).on("onStart", function (task) {
            self.runningMap[task.id] = task;
            self.emit("onStart", [task]);
        }).on("onWait", function (task) {
            delete self.runningMap[task.id];
            self.emit("onWait", [task]);
            self.runTask();
        }).on("onPause", function (task) {
            delete self.runningMap[task.id];
            self.emit("onPause", [task]);
            self.runTask();
        }).on("failed", function (task) {
            delete self.runningMap[task.id];
            self.emit("failed", [task]);
            self.runTask();
        }).on("onAbort", function (task) {
            delete self.runningMap[task.id];
            self.emit("onAbort", [task]);
            self.runTask();
        }).on("end", function (task) {
            delete self.runningMap[task.id];
            self.emit("end", [task]);
            self.runTask();
        });
};

TaskManager.prototype.createDownloadPool = function () {
    var self = this;
    //初始化下载池
    self.downloadPool = new DownloadPool(self.s3, db);
    self.downloadPool.on("progress", function (task, info) {
        self.emit("progress", [task, info]);
    }).on("onStart", function (task) {
        self.runningMap[task.id] = task;
        self.emit("onStart", [task]);
    }).on("onWait", function (task) {
        delete self.runningMap[task.id];
        self.emit("onWait", [task]);
        self.runTask();   //delete掉runningmap的项后都应该触发新任务创建
    }).on("onPause", function (task) {
        delete self.runningMap[task.id];
        self.emit("onPause", [task]);
        self.runTask();
    }).on("failed", function (task) {
        delete self.runningMap[task.id];
        self.emit("failed", [task]);
        self.runTask();
    }).on("onAbort", function (task) {
        delete self.runningMap[task.id];
        self.emit("onAbort", [task]);
        self.runTask();
    }).on("end", function (task) {
        delete self.runningMap[task.id];
        self.emit("end", [task]);
        self.runTask();
    });
}

//设置上传参数
TaskManager.prototype.setUploadConfig = function setUploadConfig(uploadConfig) {
    this.uploadConfig = uploadConfig;
    if (uploadConfig.maxUploadTasks) {
        this.maxTaskNum = uploadConfig.maxUploadTasks;
        for (let index = 0; index < this.maxTaskNum; index++) {
            this.runTask();
        }
    }
};

//设置下载参数
TaskManager.prototype.setDownloadConfig = function setDownloadConfig(downloadConfig) {
    if (downloadConfig) {
        this.downloadConfig = downloadConfig;
    }
};

//query任务数
TaskManager.prototype.countTasks = function (query, callback) {
    var self = this;
    query.userId = self.userId;
    db.count(query, callback);
};

//query任务列表
TaskManager.prototype.listTasks = function (query, pageSize, pageNum, callback) {
    var self = this;
    query.userId = self.userId;
    db.list(query, pageSize, pageNum, callback);
};
//等待任务列表
TaskManager.prototype.listWaitTasks = function (pageSize, pageNum, callback) {
    var self = this;
    db.list({userId: self.userId, stat: "WAIT"}, pageSize, pageNum, callback);
};
//暂停任务列表
TaskManager.prototype.listPauseTasks = function (pageSize, pageNum, callback) {
    var self = this;
    db.list({userId: self.userId, stat: "PAUSE"}, pageSize, pageNum, callback);

};
//失败任务列表
TaskManager.prototype.listFailedTasks = function (pageSize, pageNum, callback) {
    var self = this;
    db.list({userId: self.userId, stat: "FAILED"}, pageSize, pageNum, callback);
};
//所有任务列表
TaskManager.prototype.listAllTasks = function (pageSize, pageNum, callback) {
    var self = this;
    db.list({userId: self.userId}, pageSize, pageNum, callback);
};

//所有任务列表
TaskManager.prototype.listAllUserTasks = function (callback) {
    db.lists({}, callback);
};

//关闭DB
TaskManager.prototype.closeDB = function (callback) {
    clearInterval(this.time);
    db.close(callback);
};


//清空DB
TaskManager.prototype.clearDB = function (callback) {
    db.clear();
};

//模糊查询校验文件夹是否可删除
TaskManager.prototype.deletedCheckQuery = function (queryObj, callback) {
    var self = this;
    queryObj.userId = self.userId;
    db.regexQuery(queryObj, function (err, data) {
        if (!err) {
        }
        if (callback)
            callback(err, data);
    });
};

//批量删除任务
TaskManager.prototype.deleteTasks = function (ids, callback) {
    db.delSomeTasks(ids, callback);
};

//列举所有未完成的恢复任务
TaskManager.prototype.listAllRunRecoverTasks = function (stat, pageSize, pageNum, callback) {
    var self = this;
    var query = {$or:[{userId: self.userId, type: "recoverAndDownload", stat: stat},{userId: self.userId, type: "recover", stat: stat}]}
    db.list(query, pageSize, pageNum, callback);
};

//批量更新任务
TaskManager.prototype.updatesRecoverTasks = function (ids, callback) {
    var self = this;
    var queryString = {userId: self.userId, id: {"$in": ids}};
    var param = {$set: {stat: "SUCCESS", endTime: new Date()}};
    db.updateProperties(queryString, param, function (err, data) {
        if (callback)
            callback(err, data);
    });
};

//模糊查询校验文件夹是否可删除
TaskManager.prototype.listAndGetResult = function (queryObj, result, callback) {
    var self = this;
    queryObj.userId = self.userId;
    db.listAndGetResult(queryObj, result, function (err, data) {
        if (!err) {
        }
        if (callback)
            callback(err, data);
    });
};

function isInUploadPool(task) {
    return [Task.TaskTypeEnum.PUT, Task.TaskTypeEnum.RENAME, Task.TaskTypeEnum.COPY, Task.TaskTypeEnum.MOVE].indexOf(task.type) >= 0;
}

AWS.util.mixin(TaskManager, AWS.SequentialExecutor);
module.exports = TaskManager;