var crypto = require('crypto');
var fs = require("fs");
var _guid = require("./obsUtils").guid;

/**
 * 由于path可能含有特殊字符，所以在这里做一个处理
 * 如果文件夹或者文件最后是.，统统改为$
 * 如果路径名中除了盘符外有:,也改为$
 *
 * @param path
 */
function dealPath(path) {
    function dealColon(path) {
        var prefix = path.substring(0, 2);  //盘符
        var content = path.substring(2);
        if (content) {
            content = content.replace(/:/g, "$");
        } else {
            content = "";
        }
        return prefix + content;
    }

    function dealDot(path) {
        if (path.endsWith(".")) {
            path = path.substring(0, path.length - 1) + "$"
        }
        path = path.replace(/\.\\/g, "$\\");
        path = path.replace(/\.\//g, "$/");
        return path;
    }

    if (!path) {
        return path;
    }
    path = dealColon(path);
    path = dealDot(path);
    return path;
}


const TaskTypeEnum = function () {
    var _TaskTypeEnum = {
        PUT: "putObj",
        GET: "getObj",
        COPY: "copy",
        MOVE: "move",
        RENAME: "rename"
    };

    Object.freeze(_TaskTypeEnum);
    return _TaskTypeEnum;
}();

const TaskStatEnum = function () {
    var _TaskStatEnum = {
        WAIT: "WAIT",
        SUCCESS: "SUCCESS",
        PAUSE: "PAUSE",
        MD5Check: "MD5Check",
        MD5SUCCESS: "MD5SUCCESS",
        FAILED: "FAILED",
        MD5FAILED: "MD5FAILED",
        ABORT: "ABORT",
        SKIP:"skip"
    };
    Object.freeze(_TaskStatEnum);
    return _TaskStatEnum
}();


const DEFAULT_PART_SIZE = 5;  //字节 默认分段大小
const DEFAULT_QUEUE_SIZE = 4;  //  默认最大段数

/**
 *
 * task 对象构造函数
 * @constructor
 */
function Task({
                  key,
                  type,
                  bucket,
                  path,
                  stat = TaskStatEnum.WAIT,
                  body = undefined,
                  size = undefined,
                  id = _guid(),
                  etag = undefined,
                  errMessage = undefined,
                  storagePolicy = undefined,
                  fileName = undefined,
                  movePath = undefined,
                  partSize = DEFAULT_PART_SIZE,
                  queueSize = DEFAULT_QUEUE_SIZE,
                  userId = undefined,
                  _id = undefined,
                  startTime = new Date(),
                  endTime = undefined,
                  progress = undefined,
                  compare = undefined,
                  completeMap = undefined,
                  active = undefined,
                  uploadACL = undefined,
                  serverEncryption = undefined,
                  totalUploaded = undefined,
                  UploadId = undefined,
                  speed = undefined,
                  ContentType = undefined,
                  spTime = undefined,
                  sourceBucket = undefined,
                  sourceKey = undefined,
                  modifyTime = undefined,
                  days = undefined,
                  tier = undefined,
              }) {
    //公有属性
    this.id = id;  //..保持兼容性
    this._id = _id;
    this.userId = userId;   //创建task的userid
    this.key = key;      //远程对象key值
    this.type = type;
    this.bucket = bucket;
    this.stat = stat;
    this.body = body;
    this.path = path;    //本地路径
    this.size = size;   //byte
    this.endTime = endTime;
    this.errMessage = errMessage;      //错误信息
    this.startTime = startTime;
    this.etag = etag;

    //拷贝对象用
    this.sourceBucket = sourceBucket;
    this.sourceKey = sourceKey;
    this.modifyTime = modifyTime;

    //恢复对象用
    this.days = days;
    this.tier = tier;


    //上传下载用
    this.speed = speed;  //执行速度
    this.spTime = spTime;   //剩余时间
    this.progress = progress;     //进度
    this.partSize = partSize;           //分段大小 //单位为MB ,目前此分段一旦确定不能改变
    this.queueSize = queueSize;                       //同时分段数
    this.completeMap = completeMap;                   //已完成的段

    //上传专用属性
    this.serverEncryption = serverEncryption;  //服务器端加密方式
    this.UploadId = UploadId;                           //id，分段的时候用
    this.totalUploaded = totalUploaded;                            //总上传数
    this.storagePolicy = storagePolicy;            //存储类型
    this.uploadACL = uploadACL;    //上传对象ACL
    this.fileName = fileName;
    this.movePath = movePath;
    this.ContentType = ContentType ;
    this.compare = compare ;

    this.active = active;   //这个属性我也不知道是干啥的。在obs里面有赋值。先放着吧

    //对下载路径的特殊字符串做出处理
    if (this.type === TaskTypeEnum.GET) {
        this.path = dealPath(this.path).trim();//下载时去空格
    }
    this.computePartSize();
}


//模拟静态常量
Task.TaskStatEnum = TaskStatEnum;
Task.TaskTypeEnum = TaskTypeEnum;

Task.prototype.constructor = Task;

/**
 * 返回整体分段数
 */
Task.prototype.getTotalPartsNum = function () {
    return this.size / this.partSize / 1024 / 1024;
}

/**
 * 计算分段数和分段大小。如果按照传入的分段大小，分段数大于10000，则需要进行调整
 */
Task.prototype.computePartSize = function () {
    if (!this.size) {
        return;
    }
    let maxPartSize = Math.ceil(this.size / 10000 / 1024 / 1024);
    if (maxPartSize > this.partSize) {
        logger.info("partSize change !. task.partSize,new", this.partSize, maxPartSize)
        this.partSize = maxPartSize;
    }
}

//状态切换用的状态机 ,供changeStat使用
const TaskStatMachine = {
    "SUCCESS": {
        "ignores": ["active", "download", "req", "speed", "spTime", "completeMap", "upload"],
        "endTime": true,
        "domain": {
            "progress": "100%"
        }
    },
    "MD5SUCCESS": {
        "ignores": ["active", "speed", "spTime", "completeMap", "download", "req", "upload"],
        "endTime": true,
        "domain": {
            "progress": "100%"
        }
    },
    "skip": {
        "ignores": ["req",  "upload"],
        "endTime": true,
    },
    "MD5Check": {"ignores": ["speed", "spTime", "download", "req", "upload"], "endTime": true,},
    "MD5FAILED": {
        "ignores": ["speed", "spTime", "active", "download", "req", "upload"],
        "endTime": true,
        "domain": {
            "errMessage": "BadDigest"
        }
    },
    "PAUSE": {"ignores": ["active", "speed", "spTime"]},
    "WAIT": {"ignores": ["speed", "spTime", "errMessage"], "endTime": false},

    "FAILED": {"ignores": ["speed", "spTime"], "endTime": true}
}
/**
 * 修改task的状态。注意abort和error应该调用markerror方法。
 * @param  stat  TaskStatEnum
 * @returns {Task}
 */
Task.prototype.changeStat = function (stat) {
    var content = TaskStatMachine[stat]
    if (content.ignores) {
        for (var index = 0; index < content.ignores.length; index++) {
            this[content.ignores[index]] = undefined;
        }
    }
    if (content.endTime) {
        this.endTime = new Date();
    } else {
        this.endTime = undefined;  //以前没这个分支，这里是否会引入问题？
    }
    if (content.domain) {
        for (let i in content.domain) {
            this[i] = content.domain[i];
        }
    }
    this.stat = stat;
    return this;
}

/**
 * 根据错误对象来标识task状态
 * @param err  错误对象 或者错误字符串
 */
Task.prototype.markError = function (err) {
    if (typeof(err) == "string") {
        this.errMessage = err;
    } else {
        this.errMessage = err.code || err.message;
    }
    if (this.errMessage === "RequestAbortedError") {
        this.completeMap = undefined;
        this.totalUploaded = undefined;
        if (this.stat !== TaskStatEnum.PAUSE) {   //如果已是暂停状态，按暂停处理,这是由于非分段操作不支持暂停
            this.stat = TaskStatEnum.ABORT
        }
    } else if (this.errMessage === "RequestPausedError") {
        this.stat = TaskStatEnum.PAUSE;
    }
    else if (this.errMessage === "NoSuchUpload") { //分段上传过程中上传id被删掉了
        global.logger.warn('[NoSuchUpload]', this.UploadId, err);
        this.UploadId = undefined; //删除uploadId导致已经上传的段无效了。
        this.stat = TaskStatEnum.WAIT;   //设置为wait后重新上传
    } else {
        this.stat = TaskStatEnum.FAILED;
    }
}

/**
 * 构建上传和下载的parameter参数
 * @returns {{Bucket: *, Key: *, localpath: *, size: *}}
 */
Task.prototype.buildParam = function () {
    return {Bucket: this.bucket, Key: this.key, localpath: this.path, size: this.size};
}


/**
 * 构建下载的option参数
 * @returns {{partSize: number, queueSize: *}}
 */
Task.prototype.buildDownloadOption = function () {

    var options = {partSize: this.partSize * 1024 * 1024, queueSize: this.queueSize};
    if (this.completeMap) {
        options.completeMap = this.completeMap;
    }
    return options;
}

/**
 * 创建上传用参数
 * @param body   文件内容
 * @returns {{Bucket: *, Key: *, StorageClass: (undefined|*), ACL: *}}
 */
Task.prototype.buildUploadParam = function (body) {
    var params = {
        Bucket: this.bucket,
        Key: this.key,
        StorageClass: this.storagePolicy,
        ACL: this.uploadACL,
        ContentType:this.ContentType
    };
    if (body) {
        params.Body = body;
    } else if (body === undefined) {   //如果body没有传，则默认为对象本地path
        params.Body = this.path;
    } else {
        //如果body为其他值（0，false），表示创建空对象，不传body
    }
    if (this.serverEncryption) {
        params.ServerSideEncryption = 'aws:kms';
    }
    return params;
}

/**
 * 根据task构建上传option
 */
Task.prototype.buildUploadOptions = function () {
    var options = {
        partSize: this.partSize * 1024 * 1024,
        queueSize: this.queueSize,
        leavePartsOnError: true,
        completeMap: {},
        totalUploaded: this.totalUploaded,
    };

    if (this.UploadId) {
        options.UploadId = this.UploadId;
        options.completeMap = this.completeMap;
        options.totalUploaded = this.totalUploaded;
    }
    return options;
}

/**
 * 保存任务执行的信息，包括上传和下载
 * @param serviceHandle  上传或者下载对象，包含了分段任务的进度信息
 */
Task.prototype.saveProcessInfo = function (serviceHandle) {
    if (serviceHandle.completeMap) {    //上传和下载
        this.completeMap = serviceHandle.completeMap;
    }
    if (serviceHandle.totalUploaded) {   //上传专用
        this.totalUploaded = serviceHandle.totalUploaded;
    }
}

Task.prototype.buildHeadObjectParam = function () {
    return {
        Bucket: this.bucket,
        Key: this.key
    };
}

/**
 * 构建替换对象，主要用于设置对象元数据
 * @param metaData
 * @returns {*}
 */
Task.prototype.buildReplaceObjectParam = function (metaData) {
    var params = {
        Bucket: this.bucket,
        Key: this.key,
        MetadataDirective: "REPLACE",
        CopySource: this.bucket + "/" + encodeURIComponent(this.key),
        ACL: this.uploadACL
    }
    return Object.assign(params, metaData);
}

/**
 * 计算task中文件的md5值
 * @param callback (error,md5,size,task), 本回调为必须参数
 */
Task.prototype.computeMD5 = function (callback) {
    var self = this;
    let md5num = crypto.createHash('md5');
    let size = 0;
    fs.createReadStream(self.path).on('data', function (chunks) {  //输入流是默认读完关闭的
        md5num.update(chunks);
        size += chunks.length;
    }).on('end', function () {
        callback(null, md5num.digest("hex"), size, self);
    }).on("error", (error) => {
        callback(error, "", size, self);
    });

}

module.exports.Task = Task;