/**
 * 桶策略
 * {
"Statement" : [{
     "Sid": "ExampleStatementID1",
     "Effect": "Allow",
     "Action": "ListBucket",
     "Resource": "examplebucket"，
     "Condition": "some conditions"
  },
  {
     "Sid": "ExampleStatementID2",
     "Effect": "Allow",
     "Action": "PutObject",
     "Resource": "examplebucket"，
     "Condition": "some conditions"
  },
......
]
}
 */
guid = require("../../obsUtils").guid;

const READ_ONLY = ["GetObject", "GetObjectVersion","GetObjectAcl"];
const READ_WRITE = ["GetObject", "GetObjectVersion", "PutObject", "DeleteObjectVersion", "DeleteObject","GetObjectAcl","PutObjectAcl"]

const AWS_RESOURCE_PREFIX = "arn:aws:s3:::";
const AWS_PRINCIPAL_PREFIX = "arn:aws:iam::"
const AWS_ACTION_PREFIX = "s3:";

function isFolder(path) {
    return path.endsWith("/");
}

/**
 * 对象策略,通过桶策略解析得到, 用于对某个资源的配置简易访问策略。
 * 如果对象为文件夹，则配置的策略为文件夹及其下所有子文件夹和文件
 * 如果对象为一般对象，则配置的策略影响文件本身
 *
 * @key  对象的桶的key值，包括桶名。 如果是文件夹，需要在后面加/   (桶名/prefix/name/)
 * @param content    对象所在的桶策略
 * @constructor
 */
function ObjectPolicy(key, content) {

    this.key = key;   //传入的对象路径（包括桶名） ，文件夹需要加/,
    this.bucketName = this.key.split("/")[0];
    this.Resource = convertResourceStr(this.key);
    this.statements = [];   //解析到的关于本对象的简易策略
    //关于用户的deny策略 ,必须要给用户加入相应的deny策略
    this.denyPolicies = [];

    if (!content) {
        this.content = {};
        return;
    }
    if (typeof(content) === "string") {
        content = JSON.parse(content);
    }

    this.content = Object.assign({}, content)
    this._loadObjectPolicy();
}


/**
 *  对当前已配置的对象策略进行保存
 * @returns {string}  返回保存后的桶策略字符串
 */
ObjectPolicy.prototype.save = function () {
    if (!this.content || JSON.stringify(this.content) === "{}") { //如果之前没有任何桶策略
        if (this.statements.length === 0 && this.denyPolicies.length === 0) {
            return "";
        }
        else {
            this.content = createBucketPolicy();
        }
    }
    if (this.content.Statement.length === 0 && this.statements.length === 0 && this.denyPolicies.length === 0) {
        return "";
    }

    for (let statement of this.statements) {
        this._addToOriginalPolicy(statement);
    }
    this.denyPoliciesToBucketPolicy();
    return JSON.stringify(this.content);
}


ObjectPolicy.prototype._addToOriginalPolicy = function (simplePolicy) {
    this.content.Statement.push(simplePolicy.toS3Policy());
    this.content.Statement.push(simplePolicy.getPrefixPolicy(this.bucketName));
}

/**
 * 删除一个策略
 * @param Sid
 */
ObjectPolicy.prototype.deleteSimplePolicy = function (Sid) {
    if (!Sid) {
        logger.error("Sid cannot be null.", this.key);
        return;
    }
    for (let index in this.statements) {
        if (this.statements[index].Sid === Sid) {
            logger.info("Delete statement!:", this.key, Sid);
            let userStrs = _getUserStrFromPrincipal(this.statements[index].Principal);
            for (let userStr of userStrs) {
                this.removeDenyCondition(userStr);
            }
            remove(this.statements, index);
            break;
        }
    }
}


/**
 * 给对象策略中的涉及到的每个用户都配上denypolicy
 * 把相关的denypolicy保存到桶策略中
 */
ObjectPolicy.prototype.denyPoliciesToBucketPolicy = function () {
    for (let statement of this.statements) {
        let userStrs = _getUserStrFromPrincipal(statement.Principal);
        for (let userStr of userStrs) {
            this.addDenyCondition(userStr);
        }
    }
    this.content.Statement = this.content.Statement.concat(this.denyPolicies);
}

/**
 * 为userId增加一个本对象的deny condition
 * @param userID
 */
ObjectPolicy.prototype.addDenyCondition = function (userID) {
    for (let denyStatement of this.denyPolicies) {
        if (_isStatementInvolveUserID(denyStatement, userID)) {
            this.addConditionToDeny(denyStatement);
            return;
        }
    }
    this.denyPolicies.push(this._createS3DenyPolicy(userID));
}

ObjectPolicy.prototype.removeDenyCondition = function (userID) {
    for (let index in this.denyPolicies) {
        denyStatement = this.denyPolicies[index];
        if (_isStatementInvolveUserID(denyStatement, userID)) {
            this.removeConditionToDeny(denyStatement);
            if (!denyStatement.Condition) {
                remove(this.denyPolicies, index);
            }
            return;
        }
    }
}

/**
 * 增加本对象的前缀到deny策略的条件中
 * @param denyStatement
 */
ObjectPolicy.prototype.addConditionToDeny = function (denyStatement) {
    if (!denyStatement.Condition) {
        denyStatement.Condition = {};
    }
    if (!denyStatement.Condition.StringNotLike) {
        denyStatement.Condition.StringNotLike = {};
    }
    if (!denyStatement.Condition.StringNotLike["s3:prefix"]) {
        denyStatement.Condition.StringNotLike["s3:prefix"] = [];
    }
    let conditionPrefix = denyStatement.Condition.StringNotLike["s3:prefix"];
    let targetPrefix = convertConditionArray(this.Resource, this.bucketName);
        targetPrefix.forEach(function (data) {
        if(conditionPrefix.indexOf(data)>=0){
            return;
        }
        conditionPrefix.push(data);
    });
}

/**
 * 在deny策略的Condition中删除本对象的前缀
 * 如果删除的Condition最终为空，则需要删除整个deny策略
 * @param denyStatement
 */
ObjectPolicy.prototype.removeConditionToDeny = function (denyStatement) {
    let conditionPrefix = denyStatement.Condition &&
        denyStatement.Condition.StringNotLike && denyStatement.Condition.StringNotLike["s3:prefix"];
    if (!conditionPrefix) {   //已经移除了
        return;
    }
    let targetPrefix = convertConditionArray(this.Resource, this.bucketName);
    targetPrefix.forEach(function (data) {
        let index = conditionPrefix.indexOf(data);
        if (index >= 0) {
            remove(conditionPrefix, index);
        }
    });
    if (conditionPrefix.length == 0) {
        delete denyStatement.Condition.StringNotLike["s3:prefix"];
    }
    if (JSON.stringify(denyStatement.Condition.StringNotLike) === "{}") {
        delete denyStatement.Condition.StringNotLike;
    }
    if (JSON.stringify(denyStatement.Condition) === "{}") {
        delete denyStatement.Condition;
    }
    return;
}


/**
 * 创建一个新的deny权限,只有s3格式的
 * @param userStr  S3格式的principal字符串
 * @returns {{Sid: *, Effect: string, Action: string, Principal: *[], Resource: *[], Condition: {StringNotLike: {"s3:prefix": Array}}}}
 * @private
 */
ObjectPolicy.prototype._createS3DenyPolicy = function (userStr) {
    var denyPolicy = {
        "Sid": _buildSid(),
        "Effect": "Deny",
        "Action": ["s3:ListBucket"],
        "Principal": {"AWS": [AWS_PRINCIPAL_PREFIX + userStr]},
        "Resource": [AWS_RESOURCE_PREFIX + this.bucketName],
        "Condition": {"StringNotLike": {"s3:prefix": convertConditionArray(this.Resource, this.bucketName)}}
    };
    return denyPolicy;

}

/**
 * 增加一个简易权限
 * @Sid   规则的Sid，如果为空，则表示新建一个规则
 * @param simpleAction  只能是r或者rw字符串，否则直接返回
 * @param userInfoes       授权用户数组，同iamConnector中返回的用户对象数组
 */
ObjectPolicy.prototype.addSimplePolicy = function (Sid, simpleAction, userInfoes) {
    var Action;
    if (simpleAction === "r") {
        Action = Array.from(READ_ONLY);
    } else if (simpleAction === "rw") {
        Action = Array.from(READ_WRITE);
    }
    else {
        logger.error("[simple policy]simpleaction is invalid.", simpleAction);
        return false;
    }

    var Principal = {"ID": []};
    userInfoes.forEach(function (userInfo) {
        Principal["ID"].push("domain/" + userInfo['domain_id'] + ":user/" + userInfo['id']);
    });

    var simplePolicy = new SimplePolicy({
        "Sid": Sid,
        "Resource": this.Resource,
        "Action": Action,
        "Principal": Principal
    });
    if (Sid) {  //编辑
        for (let index in this.statements) {
            let current = this.statements[index];
            if (current.Sid === Sid) {
                this.deleteMissUserDeny(current,simplePolicy);
                this.statements[index] = simplePolicy;
                return true;
            }
        }
    } else {  //新建。不可创建同样用户
        for (let item of this.statements) {
            if (compareArrayValue(item.Principal['ID'], simplePolicy.Principal['ID'])) {
                logger.info("cannot add same user principal.", simplePolicy.Principal);
                return false;
            }
        }
        simplePolicy.Sid = _buildSid();
    }
    this.statements.push(simplePolicy);
    return true;
}

ObjectPolicy.prototype.deleteMissUserDeny = function (originalStatement, newStatement) {
    var userStr_old = _getUserStrFromPrincipal(originalStatement.Principal);
    var userStr_new = _getUserStrFromPrincipal(newStatement.Principal);
    var self=this;
    userStr_old.forEach(function(data){
        if(userStr_new.indexOf(data)<0){
            self.removeDenyCondition(data);
        }
    });
}


/**
 * 根据参数resource，获取桶策略中关于此resource的相关策略并进行解析
 *
 */
ObjectPolicy.prototype._loadObjectPolicy = function () {
    this.statements = [];  //涉及到本资源（对象）的策略配置
    let filterTemps = [];
    for (let statement of this.content.Statement) {   //过滤掉对象建议授权
        if (this._isOwnStatement(statement)) {
            this.statements.push(new SimplePolicy(statement));
        } else {
            filterTemps.push(statement);
        }
    }
    this.content.Statement = filterTemps;

    filterTemps = [];    //过滤掉前缀授权
    for (let statement of this.content.Statement) {
        if (!this._isPrefixPolicy(statement)) {
            filterTemps.push(statement);
        }
    }
    this.content.Statement = filterTemps;

    filterTemps = [];
    for (let statement of this.content.Statement) {
        if (this._isDenyPolicy(statement)) {
            this.denyPolicies.push(statement);
        } else {
            filterTemps.push(statement)
        }
    }
    this.content.Statement = filterTemps;
}

/**
 * 判断是否为简易策略的前缀策略
 * 1.Action中只能有ListBucket
 * 2.Resource中只能有当前桶
 * 3.Condition中必须为： StringLike:{"s3:prefix":[ "对象名称/*"（对象为文件夹时）, "对象名称" ]}
 * 4.Effect必须为Allow
 * 5.Principal必须能够与this.statements中一个statement的Principal保持一致，那么他就是这个statment的前缀策略
 *
 * 如果是，则返回true，并将该前缀策略与对应的简易策略关联
 * 如果不是，则返回false
 *
 * @param statement
 * @private
 */
ObjectPolicy.prototype._isPrefixPolicy = function (statement) {
    if (statement.Effect !== "Allow") {
        return false;
    }

    if (!this._checkAction(statement, "ListBucket")) {
        return false;
    }

    if (!this._checkResource(statement, this.bucketName)) {
        return false;
    }

    //这里stringlike暂不做去s3处理
    if (statement.Condition && statement.Condition.StringLike) {
        let stringLike = statement.Condition.StringLike;
        let condition = stringLike["s3:prefix"];
        if (!condition) {
            return false;
        }
        let objPrefixConditions = convertConditionArray(this.Resource, this.bucketName);
        if (!compareArrayValue(objPrefixConditions, condition)) {
            return false;
        }
    }

    var principal = _convertPrincipal(statement.Principal);
    for (let item of  this.statements) {
        if (compareArrayValue(item.Principal.ID, principal.ID)) {
            item.prefixPolicy = statement;
            return true;
        }
    }
    return false;
}

/**
 * 判断是否为denyPolicy
 * 1. Effect必须为Deny
 * 2. Action必须为ListBucket
 *
 * @param statement
 * @returns {boolean}
 * @private
 */
ObjectPolicy.prototype._isDenyPolicy = function (statement) {
    if (statement.Effect !== "Deny") {
        return false;
    }
    if (!this._checkAction(statement, "ListBucket")) {
        return false;
    }
    return true;
}


/**
 * 判断指定策略是否为针对本对象的策略,
 * 1. Resource与本对象中的Resource完全相同
 * 2. Effect为Allow
 * 3. 没有condition
 *
 * @param statement
 * @returns {boolean}
 */
ObjectPolicy.prototype._isOwnStatement = function (statement) {
    var resourceItems = statement["Resource"];

    temp = [];
    resourceItems.forEach(function (item, index) {
        temp[index] = truncateResourcePrefix(item);
    });
    if (!compareArrayValue(this.Resource, temp)) {
        return false;
    }
    if (statement.Effect !== "Allow") {
        return false;
    }
    if (statement.Condition) {
        return false;
    }
    return true;
}

/**
 * 检查statement中的action是否包含给定的action
 * @param statement
 * @param action
 * @returns {boolean}
 * @private
 */
ObjectPolicy.prototype._checkAction = function (statement, action) {
    if (statement.Action) {
        if (_convertAction(statement.Action).indexOf(action) < 0) {
            return false;
        }
    } else {
        return false;
    }
    return true;
}

/**
 * 检查statement中的resource是否为给定的resource
 * @param statement
 * @param resource
 * @private
 */
ObjectPolicy.prototype._checkResource = function (statement, resource) {
    if (statement.Resource && statement.Resource.length === 1) {
        if (_convertResource(statement.Resource)[0] !== resource) {
            return false;
        }
    } else {
        return false;
    }
    return true;
}

/**
 * 简易策略对象:简易策略对象可以简单设置只读或者读写权限
 * 如果是s3的桶策略格式，都会转化为obs格式
 *
 * @param Sid
 * @param Resource    资源名称，如果是字符串，则需要在资源名称结尾加入/已表示是否为文件夹
 * @param Action
 * @param Effect
 * @param Condition
 * @constructor
 */
function SimplePolicy({Sid = undefined, Resource = undefined, Action = [], Effect = "Allow", Principal = {}, Condition = undefined,} = {}) {
    this.Sid = Sid;
    this.Effect = Effect;
    this.Resource = _convertResource(Resource);
    this.Principal = _convertPrincipal(Principal);
    this.Action = _convertAction(Action);
    this.Condition = Condition;
    this.prefixPolicy = false;    //前缀listBucket,这个策略不做去s3处理
}


/**
 * 这里创出来的是s3格式
 * @param bucketName
 * @returns {*}
 */
SimplePolicy.prototype.getPrefixPolicy = function (bucketName) {
    if (this.prefixPolicy) {
        return this.prefixPolicy;
    }
    var result = {
        Sid: _buildSid(),
        Effect: "Allow",
        Resource: [AWS_RESOURCE_PREFIX + bucketName],
        Action: ["s3:ListBucket"],
        Principal: this.toS3Principal(),
        Condition: {"StringLike": {"s3:prefix": convertConditionArray(this.Resource, bucketName)}}
    };
    this.prefixPolicy = result;
    return result;
}

SimplePolicy.prototype.toS3Principal = function () {
    var result = {"AWS": []};
    this.Principal["ID"].forEach(function (item, index) {
        var temp = item.substring("domain/".length);
        result["AWS"][index] = AWS_PRINCIPAL_PREFIX + temp;
    });
    return result;
}

SimplePolicy.prototype.toS3Policy = function () {
    var result = {Sid: this.Sid, Effect: this.Effect, Resource: [], Action: [], Principal: {"AWS": []}};


    this.Resource.forEach(function (item, index) {
        result['Resource'][index] = AWS_RESOURCE_PREFIX + item;
    });
    this.Action.forEach(function (item, index) {
        result['Action'][index] = AWS_ACTION_PREFIX + item;
    });
    result['Principal'] = this.toS3Principal();
    return result;
}

/**
 *
 * @param userInfo  iam里面拿到的用户信息数组，数据结构见IAMConnector，queryUser回调返回
 * @returns {*}
 */
SimplePolicy.prototype.getUserInfo = function (userInfoes) {
    if (!userInfoes) {
        userInfoes = require("../../access/iam/IAMConnector").userCache;
    }
    var result = []
    for (let p of this.Principal["ID"]) {
        var user_id = p.split("user/")[1].trim();
        for (var info of userInfoes) {
            if (info.id === user_id) {
                result.push(info)
            }
        }
    }
    return result;
}

var _convertResource = function (Resource) {
    var temp = [];
    Resource.forEach(function (resourceStr, index) {
        resourceStr = truncateResourcePrefix(resourceStr);
        temp[index] = resourceStr;
    });
    return temp;
}

var _convertAction = function (Action) {
    var temp = []
    Action.forEach(function (action, index) {
        if (action.startsWith(AWS_ACTION_PREFIX)) {
            action = action.substring(AWS_ACTION_PREFIX.length);
        }
        temp[index] = action;
    });
    return temp;
}

var _convertPrincipal = function (Principal) {
    if (Principal["AWS"]) {
        var temp = [];
        Principal["AWS"].forEach(function (content, index) {
            content = "domain/" + content.substring(AWS_PRINCIPAL_PREFIX.length);
            temp[index] = content
        });
        return {"ID": temp};
    } else if (Principal["ID"]) {
        return Object.assign({}, Principal);
    } else {
        logger.warn("can not parse Prinicipal.", this.path, Principal);
        return undefined;
    }
}


/**
 * 按简易策略的Action具体内容，返回只读，读写，或者其他
 * @returns {string}   r  rw  other
 */
SimplePolicy.prototype.actionToStr = function () {
    if (compareArrayValue(this.Action, READ_ONLY)) {
        return "r";
    }
    if (compareArrayValue(this.Action, READ_WRITE)) {
        return "rw";
    }
    return "other";
}


/**
 * 比较2个数组是否值相同，忽略顺序
 */
function compareArrayValue(arrayA, arrayB) {
    return JSON.stringify(arrayA.sort()) === JSON.stringify(arrayB.sort());
}


/**
 * 把资源对象的字符串path转化为桶策略中合法的资源数组
 * @param resourceStr
 * @returns {Array}
 */
function convertResourceStr(resourceStr) {
    let resource = [];
    if (isFolder(resourceStr)) {
        resource.push(resourceStr + "*");
        // resource.push(resourceStr.substring(0, resourceStr.length - 1));
    } else {
        resource.push(resourceStr);
    }
    return resource;
}

/**
 * 把资源对象转化成的resource数组转换成前缀配置使用的condition条件数组
 * @param resourceArray
 */
function convertConditionArray(resourceArray, bucketName) {
    let objPrefixConditions = [];   //判断是否为前缀允许的条件，是本对象的Resource属性去掉桶名
    resourceArray.forEach(function (data) {
        objPrefixConditions.push(data.substring((bucketName + "/").length));
    });
    return objPrefixConditions;
}

function createBucketPolicy() {
    var policy = {
        "Version": "2008-10-17",
        "Id": "obsbrowser-policy" + guid(),
        "Statement": []
    };
    return policy;
}


/**
 * 从策略的Principal提取出principal数组
 *
 * @param Principal
 */
function _getUserStrFromPrincipal(Principal) {
    var contents = Principal['AWS'] || Principal['ID'];
    var result = [];
    contents.forEach(function (data) {
        if (data.startsWith(AWS_PRINCIPAL_PREFIX)) {
            result.push(data.substring(AWS_PRINCIPAL_PREFIX.length));
        } else {
            result.push(data.substring("domain/".length));
        }
    });
    return result;
}

function truncateResourcePrefix(resourceStr) {
    if (resourceStr.startsWith(AWS_RESOURCE_PREFIX)) {
        resourceStr = resourceStr.substring(AWS_RESOURCE_PREFIX.length);
    }
    return resourceStr;
}

/**
 * statement的权限配置是否涉及到指定的userid
 * @param statment
 * @param userID
 * @returns {boolean}
 * @private
 */
function _isStatementInvolveUserID(statement, userID) {
    var users = statement.Principal.AWS || statement.Principal.ID;
    for(let user of users){
        if (user.endsWith(userID)) {
            return true;
        }
    }
    return false;
}


function _buildSid() {
    return "obsbrowser-object" + guid();
}

// Array Remove - By John Resig (MIT Licensed)
var remove = function (arrays, from, to) {
    from=parseInt(from);
    var rest = arrays.slice((to || from) + 1 || arrays.length);
    arrays.length = from < 0 ? arrays.length + from : from;
    return arrays.push.apply(arrays, rest);
};


module.exports = ObjectPolicy;